from tqdm import tqdm
from typing import Optional, Dict, Union
from isahitlab.operations.base import BaseAction
from isahitlab.domain.batch import BatchId
from isahitlab.domain.dataset import DatasetId
from isahitlab.api.batch.api import BatchApi

from typeguard import typechecked

class LinkDatasetToBatchOperation(BaseAction):
    """Link (or unlink) a dataset to a batch operation"""

    @typechecked
    def run(
        self,
        batch_id: BatchId,
        dataset_id : Union[DatasetId, None],
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Link (or unlink) a dataset to a batch
        
        Args:
            batch_id: ID of the batch
            dataset_id : ID of the dataset to link or None to unlink a dataset from the batch
            disable_progress_bar: Disable the progress bar display
        """

        batch_api = BatchApi(self._http_client)
        with tqdm(total=1,  disable=disable_progress_bar, desc="Updating batch... ") as loader:
            updated_batch = batch_api.link_dataset_to_batch(batch_id=batch_id, dataset_id=dataset_id)
            loader.update(1)

        return updated_batch
    


from typing import Dict, Generator, Iterable, Optional

from isahitlab.api.project_configuration.api import ProjectConfigurationApi
from isahitlab.api.task.api import TaskApi
from isahitlab.domain.project import ProjectId
from isahitlab.domain.task import TaskCompatibilityMode, TaskFilters
from isahitlab.formatters import get_compatibility_formatter
from isahitlab.operations.base import BaseAction
from tqdm import tqdm
from typeguard import typechecked


class GetTasksOperation(BaseAction):
    """Tasks actions"""

    @typechecked
    def run(
        self,
        project_id: ProjectId,
        filters: TaskFilters,
        disable_progress_bar: Optional[bool] = False,
        raw_data: Optional[bool] = False,
        compatibility_mode: Optional[TaskCompatibilityMode] = None
    ) -> Generator[Iterable[Dict],None,None]:
        """ Get a tasks list generator
        
        Args:
            project_id: ID of the project
            filters : TaskFilters object
            disable_progress_bar: Disable the progress bar display
            raw_data: Get raw data and metadata object instead of merged data
            compatibility_mode: Format the output for specific use cases
                Possible choices: `kili` -> format the ouput to look like kili.assets() results 

        """
        task_api = TaskApi(self._http_client)

        # Load configuration and initialize the formatter if required
        compatibility_mode = compatibility_mode or ("lab" if not raw_data else None)

        formatter = None
        if compatibility_mode:
            project_configuration_api = ProjectConfigurationApi(self._http_client)
            with tqdm(total=1,  disable=disable_progress_bar, desc="Loading project configuration... ") as loader:
                project_configuration = project_configuration_api.get_project_configuration(project_id)
                loader.update(1)            
            formatter = get_compatibility_formatter("lab", compatibility_mode, project_configuration)

        # load and format task
        with tqdm(total=0,  disable=disable_progress_bar, desc="Loading tasks... ") as loader:
            for (docs, loaded, total) in task_api.get_all_tasks(filters):
                loader.total = total
                if formatter :
                    docs = formatter.format_tasks(docs)
                
                yield from docs
                loader.update(loaded - loader.n)

    



from typing import Any, Dict

from isahitlab.api.project_configuration.api import ProjectConfigurationApi
from isahitlab.api.task.api import TaskApi
from isahitlab.domain.export import ExportFormat
from isahitlab.domain.project import ProjectId
from isahitlab.domain.task import TaskFilters
from isahitlab.formatters import get_export_formatter
from isahitlab.operations.base import BaseAction
from tqdm import tqdm
from typeguard import typechecked


class ExportOperation(BaseAction):
    """Tasks actions"""

    @typechecked
    def run(
        self,
        project_id: ProjectId,
        format: ExportFormat,
        filters: TaskFilters,
        options: Dict = {},
        disable_progress_bar: bool = False
    ) -> Any :
        """ Export tasks
        
        Args:
            project_id: ID of the project
            filters : TaskFilters object
            disable_progress_bar: Disable the progress bar display
            compatibility_mode: Format the output for specific use cases
                Possible choices: `kili` -> format the ouput to look like kili.assets() results 

        """
        project_configuration_api = ProjectConfigurationApi(self._http_client)
        task_api = TaskApi(self._http_client)
       

        # Load configuration and initialize the formatter if required
        with tqdm(total=1,  disable=disable_progress_bar, desc="Loading project configuration... ") as loader:
            project_configuration = project_configuration_api.get_project_configuration(project_id)
            loader.update(1)            

        formatter = get_export_formatter("lab", format, project_configuration, { **options, "extra_filename_part" : project_id  })

        if not formatter:
            raise Exception("No formatter found for {}".format(format))
        
        # load and format task
        with tqdm(total=0,  disable=disable_progress_bar, desc="Loading tasks... ") as loader:
            for (docs, loaded, total) in task_api.get_all_tasks(filters):
                loader.total = total
                for task in docs:
                    formatter.append_task_to_export(task)
                loader.update(loaded - loader.n)

        return formatter.complete_export()
    



import json
import logging
import pathlib
from datetime import datetime
from typing import Any, Dict, List, Union

from isahitlab.helpers.data import merge_data_and_metadata
from isahitlab.helpers.labels import extract_labels_map_by_id, extract_main_labels
from isahitlab.helpers.path import check_extension

from .base import BaseFormatter

logger = logging.getLogger(__name__)

class LabToLabFormatter(BaseFormatter):
    """Formatter used to merge task data and metadata"""

    is_sequence: bool = False

    export : List[Dict[str, Any]]= None

    def __init__(self, project_configuration : Dict, options : Dict = {}) -> None :

        self._init_configuration(project_configuration, options, {
        })
        
        # Check output filename
        if options.get("output_filename", None) and not check_extension(options.get("output_filename", ""), ".json"):
            raise Exception('"output_filename" must have .json extension')
        
        self.is_sequence = self._project_configuration.get("metadata", {}).get("toolOptions", {}).get("sequence", False)
        

    ### Transform

    def format_tasks(self, tasks: List[Dict]) -> List[Dict]:
        formatted_tasks = []

        for task in tasks:
            formatted_task = self._format_task(task)
            formatted_tasks.append(formatted_task)

        return formatted_tasks
    
    ### Import
    def prepare_import(self, tasks):
        formatted_tasks = []
        labels = None
        ## If is IAT
        if self.is_iat_project_type():
            labels = extract_labels_map_by_id(self._project_configuration)
        for task in tasks:
            formatted_tasks.append(self._prepare_task(task, labels))
        return formatted_tasks
    
    def _prepare_task(self, task: Dict, labels : List = None) -> Dict:
        data = task.get("data", None)
        if data != None:
            prepared_data = {
                **data
            }
            ## If is IAT
            if self.is_iat_project_type():
                annotations = prepared_data.get("annotations",[])
                prepared_data["annotations"] = self._prepare_annotations(annotations, labels)
            elif self.is_form_project_type():
                iat_inputs = self.get_tool_iat_input()
                for iat_input in iat_inputs:
                    input_data = prepared_data.get(iat_input['id'], None)
                    if input_data == None:
                        continue
                    annotations = input_data.get("annotations",[])
                    input_labels_map = extract_labels_map_by_id(iat_input.get('config', {}))
                    prepared_data[iat_input['id']]["annotations"] = self._prepare_annotations(annotations, input_labels_map)
            task['data'] = prepared_data
        return task
        

    def _prepare_annotations(self, annotations: List, labels_list: Dict):
        prepared_annotations = []
        for annotation in annotations:
            labels = annotation.get("labels", {})
            main_labels = extract_main_labels(labels)
            texts = []
            colors = []
            
            for main_label in main_labels:
                label_config = labels_list.get(main_label['id'], None)
                if label_config:
                    texts.append(label_config['name'])
                    colors.append(label_config['color'])
            
            annotation_text = ", ".join(texts) if len(texts) > 0 else None
            annotation_color = colors[len(colors) - 1] if len(colors) > 0 else None

            if annotation_text and 'text' not in annotation:
                annotation['text'] = annotation_text
            if annotation_color and 'color' not in annotation:
                annotation['color'] = annotation_color
            
            prepared_annotations.append(annotation)

        return prepared_annotations

    ### Export

    def _init_export(self):
        self.export = []
    
    def append_task_to_export(self, task: Dict):
        if self.export == None:
            self._init_export()

        self.export.append(self._format_task(task))

    def complete_export(self) -> Any:
        if self.export == None:
            logger.warning('Nothing to export')
            return
        
        if self._options.get("in_memory"):
            return self.export

        now_str = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        default_folder = '.'
        extra_filename_part = "_{}".format(self._options.get("extra_filename_part")) if "extra_filename_part" in self._options else ""
        default_filename = f'lab{extra_filename_part}_{now_str}.json'
        output_path = pathlib.Path(self._options["output_folder"] or default_folder)
        export_name = self._options["output_filename"] or default_filename

        outputfile_path = output_path / export_name
        outputfile_path.parent.mkdir(parents=True, exist_ok=True)

        with outputfile_path.open('w') as ef:
            json.dump(self.export, ef, indent=2)

        print("Exported to \"{}\"".format(str(pathlib.Path(outputfile_path).absolute())))

    def _format_task(self, task : Dict) -> Dict:
        task = self._normalize_task(task)
        formatted_task = {
            **task,
            "data" : self._map_data_body(task)
        }

        return formatted_task


    def _map_data_body(self, task: Dict) -> Union[Dict, None]:
        if not task.get("data",None):
            return {}        
        
        data = task["data"].get("body", None)
        metadata = (task["data"].get("metadata", {}) or {}).get("body", None)
        if not data and not metadata:
            return None
        
        merged_data = merge_data_and_metadata(data, metadata)

        return merged_data
        



import logging
from typing import Dict, List, Union

from isahitlab.domain.project import ProjectType
from isahitlab.domain.task import TaskPayload
from isahitlab.helpers.labels import find_label_in_main_list
from isahitlab.helpers.points import (corners_to_bbox, normalize_points,
                                      points_to_vertices)

from .base import BaseFormatter

logger = logging.getLogger(__name__)

class KiliToLabFormatter(BaseFormatter):
    """Formatter used to ensure retrocompatibilty with kili"""

    project_configuration: Dict
    is_sequence: bool = False
    project_type: ProjectType = None

    def __init__(self,project_configuration : Dict) -> None :
        self.project_configuration = project_configuration
        
        self.is_sequence = project_configuration.get("metadata", {}).get("toolOptions", {}).get("sequence", False)
        self.project_type = project_configuration['projectType']

        if self.project_type == "form":
            raise Exception("Kili compatibiliy is not supported for form project")

        if self.is_sequence:
            raise Exception("Kili compatibiliy is not supported for sequence project")


    def format_tasks(self, tasks: List[TaskPayload]) -> List[Dict]:

        formatted_tasks = []
        for task in tasks:
            formatted_tasks.append(TaskPayload(
                task.name, 
                task.resources, 
                self._format_data(task.data),
                task.properties
                ))
        
        return formatted_tasks

    def append_task_to_export(self, task: Dict):
        raise NotImplementedError
    
    def complete_export(self, tasks: List[Dict]):
        raise NotImplementedError

    def _format_data(self, data : Dict) -> Dict:
        formatted_data = {
            "annotations": []
        }

        annotations = data["OBJECT_DETECTION_JOB"]["annotations"]
        resolution = data.get("resolution", None)

        for annotation in annotations:
            formatted_data["annotations"].append(self._format_annotation(annotation, resolution))


        return formatted_data

    def _format_annotation(self, annotation : Dict, resolution : Union[Dict, None]) -> Dict:

        formatted = {
            "polygons": [],
            "labels" : {}
        }

        bounding_polys = annotation["boundingPoly"]

        for bounding_poly in bounding_polys:
            if self.project_type == 'iat-rectangle':
                formatted["polygons"].append(self._format_bounding_poly_for_rectangle(bounding_poly, resolution))
            else:
                formatted["polygons"].append(self._format_bounding_poly_for_polygon(bounding_poly, resolution))

        if "categories" in annotation and len(annotation["categories"]) > 0:
            label_id = annotation["categories"][0]["name"]
            if label_id:
                matching = find_label_in_main_list(label_id, self.project_configuration)
                if not matching:
                    raise Exception("No label found for category \"{}\"".format(label_id))
                formatted["labels"] = matching

        if "children" in annotation and annotation['children'] and len(annotation['children']) > 0:
            logger.warning("Mapping of \"children\" is not supported")

        return formatted
    
    def _format_bounding_poly_for_rectangle(self, bounding_poly : Dict, resolution : Union[Dict, None]) -> Dict:
        if "normalizedVertices" in bounding_poly:
            vertices = corners_to_bbox(points_to_vertices(bounding_poly["normalizedVertices"]))
        elif "vertices" in bounding_poly:
            if not resolution:
                raise Exception("Resolution (width, height) is required to normalize vertices. You can also directly provide \"normalizedVertices\" instead of \"vertices\".")
            vertices = corners_to_bbox(points_to_vertices(normalize_points(bounding_poly["vertices"])))

        return {
            "geometry": {
                "vertices": vertices,
                "type": "rectangle"
            }
        }
    
    def _format_bounding_poly_for_polygon(self, bounding_poly : Dict, resolution : Union[Dict, None]) -> Dict:
        if "normalizedVertices" in bounding_poly:
            vertices = points_to_vertices(bounding_poly["normalizedVertices"])
        elif "vertices" in bounding_poly:
            if not resolution:
                raise Exception("Resolution (width, height) is required to normalize vertices. You can also directly provide \"normalizedVertices\" instead of \"vertices\".")
            vertices = points_to_vertices(normalize_points(bounding_poly["vertices"]))

        return {
            "geometry": {
                "vertices": vertices,
                "type": "polygon"
            }
        }


    def prepare_import(self, tasks: List[Dict]):
        raise NotImplementedError


import re

def is_url(path: object):
    """Check if the path is a url or something else.

    Args:
        path: path of the file
    """
    return isinstance(path, str) and re.match(r"^(http://|https://)", path.lower())


def is_resource(path: object):
    """Check if the path is a reference to a dataset resource.

    Args:
        path: path of the file
    """

    RESOURCE_PREFIX_STR = 'resource://'

    return isinstance(path, str) and path[0:len(RESOURCE_PREFIX_STR)] == RESOURCE_PREFIX_STR


import json
import logging
import uuid
from typing import Dict, Iterable, List, Optional, Union

from isahitlab.api.data.api import DataApi
from isahitlab.api.file.api import FileApi
from isahitlab.api.project_configuration.api import ProjectConfigurationApi
from isahitlab.api.task.api import TaskApi
from isahitlab.domain.batch import BatchId
from isahitlab.domain.project import ProjectId
from isahitlab.domain.task import TaskCompatibilityMode, TaskPayload
from isahitlab.formatters import get_compatibility_formatter, get_creation_formatter
from isahitlab.helpers.list import list_duplicates, remove_duplicates
from isahitlab.helpers.resource import is_resource, is_url
from isahitlab.operations.base import BaseAction
from isahitlab.validators import validate_task_payloads, validate_task_project_configuration_overwrite
from tqdm import tqdm
from typeguard import typechecked

logger = logging.getLogger('isahitlab.create_task')


class CreateTaskOperation(BaseAction):
    """Create task actions
    """

    _task_api : TaskApi
    _file_api : FileApi
    _data_api : DataApi

    @typechecked
    def run(
        self,
        project_id: ProjectId,
        batch_id: BatchId,
        tasks: List[TaskPayload],
        disable_progress_bar: Optional[bool] = False,
        disable_data_check: Optional[bool] = False,
        disable_unicity_check: Optional[bool] = False,
        raise_if_existing: Optional[bool] = True,
        compatibility_mode: Optional[TaskCompatibilityMode] = None
    ) -> Iterable[Dict]:
        """ Create tasks in a batch
        
        Args:
            project_id: ID of the project
            batch_id: ID of the batch project
            tasks : list of the tasks to create. See TaskPayload(name : str, resource : str, data : Dict[str, Any])
            disable_progress_bar: Disable the progress bar display
            disable_data_check: Set this option to False to ignore the data format validation
            disable_unicity_check: Set this option to True to ignore duplicate check and insert it anyway
            raise_if_existing: Set this option to False to skip task duplicates and only show a warning
            compatibility_mode: Format the output for specific use cases
                Possible choices: `kili` -> format the ouput to look like kili.assets() results 

        """
        self._task_api = TaskApi(self._http_client)
        self._file_api = FileApi(self._http_client)
        self._data_api = DataApi(self._http_client)
        project_configuration_api = ProjectConfigurationApi(self._http_client)
        
        project_configuration = None
        
        formatter = None

        if compatibility_mode:
            project_configuration = project_configuration_api.get_project_configuration(project_id=project_id) if project_configuration == None else project_configuration
            formatter = get_compatibility_formatter(compatibility_mode, "lab", project_configuration)
            validate_task_payloads(tasks, project_configuration, compatibility_mode)
            tasks = formatter.format_tasks(tasks)

        if not disable_data_check:
            # Check data against configuration
            project_configuration = project_configuration_api.get_project_configuration(project_id=project_id) if project_configuration == None else project_configuration
            validate_task_payloads(tasks, project_configuration, "lab")
            validate_task_project_configuration_overwrite(tasks, project_configuration)

        creation_formatter = get_creation_formatter("lab", project_configuration)
        if creation_formatter:
            tasks = creation_formatter.prepare_import(tasks)

        if not disable_unicity_check:
            tasks = self._check_unicity_and_filter(batch_id=batch_id,raise_if_existing=raise_if_existing,tasks=tasks)
            
        nb_tasks = len(tasks)
        created = []
        
        with tqdm(total=nb_tasks,  disable=disable_progress_bar, desc="Creating tasks... ") as loader:
            for task in tasks:
                created.append(self._create_task(batch_id, task))
                loader.update(1)
            

        
        return created
    

    def _check_unicity_and_filter(self, batch_id: BatchId, raise_if_existing: bool, tasks : List[TaskPayload]) ->  List[TaskPayload]:
        # Check unicity
        task_names = [*map(lambda t : t.name, tasks)]
        
        # Check if two or more tasks have the same name
        dupes = list_duplicates(task_names)
        nb_dupes = len(dupes)
        if nb_dupes > 0:
            raise Exception(
"""You have {} duplicate task name(s): {}{}
    Set disable_unicity_check option to True to ignore this error
""".format(str(nb_dupes), ','.join(dupes[0:3]), " and {} more".format(str(nb_dupes - 3)) if nb_dupes > 3 else '')
            )

        # Check tasks with the same name exits in the batch
        existings = remove_duplicates(self._task_api.get_existing_task_names(batch_id=batch_id, names=task_names))
        if len(existings):
            nb_existings = len(existings)
            existings_message = "{} task name(s) already exist in your batch: {}{}".format(str(nb_existings), ','.join(existings[0:3]), " and {} more".format(str(nb_existings - 3)) if nb_existings > 3 else '')
            if raise_if_existing:
                raise Exception(
"""{}
    Set raise_if_existing option to False to skip these tasks and only show a warning
    Set disable_unicity_check option to True to ignore and insert it anyway
""".format(existings_message))
            else:
                logger.debug('debug')
                logger.warning(
"""{}
    Set disable_unicity_check option to True to ignore and insert it anyway
""".format(existings_message))
                tasks = list(filter(lambda t: t.name not in existings, tasks))
        return tasks
    
    def _create_task(self, batch_id: str, task : TaskPayload) -> str:
        resources = []
        grant_resources_token = None
        data_id = None
        grant_data_token = None

        # Handle resources
        if task.resources:
            (resources, grant_resources_token) = self._create_resources(batch_id=batch_id, resources=task.resources)

        # Handle data
        if task.data:
            (data_id, grant_data_token) = self._create_data(batch_id=batch_id, name=task.name, data={
                **task.data,
                "__project_configuration_overwrite__" : task.project_configuration_overwrite
            })


        # Finally, create the taks
        return self._task_api.create_task(batch_id=batch_id, name=task.name, resources=resources, properties=task.properties, data_id=data_id, grant_resources=grant_resources_token, grant_data=grant_data_token)


    def _upload_file(self, resource: str, folder: str, batch_id: str) -> Dict:
        with open(resource, 'rb') as f:
            file_info = self._file_api.upload_file(batch_id=batch_id, file=f, folder=folder)
        return file_info


    def _create_resources(self, batch_id: str, resources: List[Union[str,Dict]]) -> tuple[List[Union[str,Dict]], str]:
        normalized_resources = []
        file_ids = []
        for resource in resources:
            if not is_url(resource) and not is_resource(resource):
                # Assume it is a local file
                folder = str(uuid.uuid4())
                file_info = self._upload_file(resource, folder, batch_id)
                file_ids.append(file_info['_id'])
                normalized_resources.append({
                    "id": file_info['_id'], 
                    "name":  file_info["name"]
                })
            else:
                normalized_resources.append(resource)
        grant_resources_token = None
        if len(file_ids):
            grant_resources_token = self._file_api.grant_file_access(ids=file_ids)["accessToken"]
        
        return (normalized_resources, grant_resources_token)
    
    def _create_data(self, batch_id: str, data: Dict, name: str) -> tuple[str, str]:
        data_result = self._data_api.import_data(batch_id=batch_id,name=name, body=data)
        data_id = data_result['initialDataId']
        grant_data_token = data_result['accessToken']
        return (data_id, grant_data_token)



LAB_IAT_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "$id": "iat",
    "properties": {
        "resources" : {
            "type" : "array",
            "items": {
                "type": "string"
            }
        },
        "annotations": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "uid": {"type": "string"},
                    "color": {"type": "string"},
                    "text": {"type": "string"},
                    "polygons": {
                        "type": "array",
                        "items":  {
                            "type": "object",
                            "properties": {
                                "id":  {"type": "string"},
                                "color":  {"type": "string"},
                                "geometry":  {
                                    "type": "object",
                                    "properties":{
                                        "vertices" : { "$ref" : "iat#/$defs/polygon_vertices" },
                                        "type" : { "enum" : ["polygon", "graph"] }
                                    },
                                    "required": ["vertices", "type"],
                                },
                                "groupId" : { "type" : "string" },
                                "text" : { "type" : "string" },
                            },
                            "required": ["geometry"],
                        }
                     },
                    "labels" : { "$ref" : "iat#/$defs/labels" }
                },
                "required": ["polygons"],
            }
        },
        "image": {
          "width": { "type" : "number" },
          "height": { "type" : "number" }
        },
    },
    "required": ["annotations"],
    "additionalProperties": False,
    "$defs": {
        "vertice": {
            "type": "number",
            "minimum": 0,
            "maximum": 1
        },
        "polygon_vertices": { 
            "type": "array", 
            "items": { "$ref" : "iat#/$defs/vertice" }
        },
        "rectangle_vertices": { 
            "type": "array", 
            "prefixItems": [
                { "$ref" : "iat#/$defs/vertice" },
                { "$ref" : "iat#/$defs/vertice" },
                { "$ref" : "iat#/$defs/vertice" },
                { "$ref" : "iat#/$defs/vertice" },
            ]
        },
        "labels": {
            "type": "object",
            "patternProperties": {
                "^[0-9a-zA-Z-_ ]+$": {
                    "type" : "object",
                    "properties":    {
                        "labels" : {
                            "type" : "array",
                            "items" : {
                                "type" : "object",
                                "properties": {
                                    "id" : { "type" : "string"},
                                    "name" : { "type" : "string"},
                                    "displayName" : {  "oneOf": [{"type" : "string"}, { "type" : "null"}]},
                                    "uid" : { "type" : "string"},
                                },
                                "required": ["id", "name"]
                            }
                        },
                        "children" : { "$ref" : "iat#/$defs/labels" }
                    },
                    "additionalProperties": False,
                }
            },
            "additionalProperties": False,
        }
    },
}


""" Domain Filter to API params"""
from isahitlab.domain.batch import BatchFilters

def batch_filters_mapper(filters: BatchFilters):
    params = {
        "pagination" : "false" if filters.pagination == False else "true",
        "page" : filters.page,
        "limit" : filters.limit,
        "projectId" : filters.project_id,
    }

    if filters.status_in and len(filters.status_in):
        params['statuses'] = filters.status_in

    if filters.search and len(filters.search):
        params['search'] = filters.search

    return params


""" Domain Filter to API params"""
from isahitlab.domain.project import ProjectFilters

def project_filters_mapper(filters: ProjectFilters):
    params = {
        "pagination" : "false" if filters.pagination == False else "true",
        "page" : filters.page,
        "limit" : filters.limit
    }

    if filters.id_in and len(filters.id_in):
        params['ids'] = filters.id_in

    if filters.status_in and len(filters.status_in):
        params['statuses'] = filters.status_in

    if filters.type_in and len(filters.type_in):
        params['types'] = filters.type_in

    if filters.search and len(filters.search):
        params['search'] = filters.search
        
    return params


from typing import NewType, Union, Tuple, Optional, TypedDict, Iterable
import io
from dataclasses import dataclass
from isahitlab.domain.integration import IntegrationId

DatasetId = NewType("DatasetId", str)

@dataclass
class FilePayload(dict):
    """Task payload for task create

    Inherit from dict to make it JSON serializable
    
    Args:
        file: Name of the task (appended to the path, it must be unique in a dataset)
        path: Path of a resource (ex.: "./images/image1.jpg")
    """

    file: Union[str, Tuple[str, io.IOBase]]
    path:  Union[str, None]
    

@dataclass
class DatasetBrowsingFilters:
    """Dataset browsing filters
    Args:
        dataset_id: ID of the dataset to browse
        folder: Path in the dataset
        next_token: Token identifying the next page
    """
    
    dataset_id: str
    folder: Optional[str] = None
    next_token: Optional[str] = None


@dataclass
class DatasetPayload:
    """Dataset payload
    
    Args:
        name: Name of the dataset
        integration_id: ID of the integration to use
        base_folder: path to a folder in the integration bucket
    """

    name: str
    integration_id:  Optional[IntegrationId] = None
    base_folder:  Optional[str] = None



class FileBrowsing(TypedDict):
    """Paginated result"""

    files: Iterable[str]
    folders: Iterable[str]
    nextToken: Union[str, None]
    



# SPDX-FileCopyrightText: 2024-present Benjamin Piogé <benjamin@isahit.com>
#
# SPDX-License-Identifier: MIT
from typing import Dict

from isahitlab.domain.export import ExportFormat
from isahitlab.domain.task import TaskCompatibilityMode

from .base import BaseFormatter
from .kili_to_lab_formatter import KiliToLabFormatter
from .lab_to_kili_formatter import LabToKiliFormatter
from .lab_to_lab_formatter import LabToLabFormatter
from .lab_to_yolo_formatter import LabToYoloFormatter
try:
    from .lab_to_mask_formatter import LabToMaskFormatter
except:
    pass


def get_compatibility_formatter(from_mode : TaskCompatibilityMode, to_mode : TaskCompatibilityMode, project_configuration : Dict) -> BaseFormatter :
    """Factory"""
    if from_mode == "lab" and to_mode == "kili":
        return LabToKiliFormatter(project_configuration)
    if from_mode == "kili" and to_mode == "lab":
        return KiliToLabFormatter(project_configuration)
    if from_mode == "lab" and to_mode == "lab":
        return LabToLabFormatter(project_configuration)

def get_creation_formatter(from_mode : TaskCompatibilityMode, project_configuration : Dict) -> BaseFormatter :
    """Factory"""
    if from_mode == "lab":
        return LabToLabFormatter(project_configuration)
    
def get_export_formatter(from_format : ExportFormat, to_format : ExportFormat, project_configuration : Dict, options: Dict = {}) -> BaseFormatter :
    """Factory"""
    if from_format == "lab" and to_format == "kili":
        return LabToKiliFormatter(project_configuration, options)
    if from_format == "lab" and to_format == "yolo":
        return LabToYoloFormatter(project_configuration, options)
    if from_format == "lab" and to_format == "lab":
        return LabToLabFormatter(project_configuration, options)
    if from_format == "lab" and to_format == "mask":
        try:
            return LabToMaskFormatter(project_configuration, options)
        except:
            raise ModuleNotFoundError(
                "OpenCV and Numpy are required to use the mask export. To install them,"
                " run: pip install isahitlab[image]."
            )
    



"""Validation for form"""

import copy
from typing import Any, Dict

from isahitlab.helpers.inputs_layout import \
  extract_not_static_inputs_from_layout

from . import validation_schemas


def build_validation_schema_for_form(project_configuration) ->  Dict[str, Any]:

    inputs_layout = project_configuration.get('metadata', {}).get('inputsLayout', None)

    if not inputs_layout:
        raise Exception('Not layout defined for the project')
    
    inputs = extract_not_static_inputs_from_layout(inputs_layout)

    schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "properties": {
        },
        "patternProperties": {
            "^_meta_": {}
        },
        "additionalProperties": False,
    }

    for input in inputs:
        schema["properties"][input["id"]] = _get_schema_for_input(input)

    return schema

def build_validation_schema_for_form_config_overwrite(project_configuration) ->  Dict[str, Any]:

    inputs_layout = project_configuration.get('metadata', {}).get('inputsLayout', None)

    if not inputs_layout:
        raise Exception('Not layout defined for the project')
    
    inputs = extract_not_static_inputs_from_layout(inputs_layout)

    schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "properties": {
        },
        "patternProperties": {
            "^_meta_": {}
        },
        "additionalProperties": False,
    }

    for input in inputs:
        # if input.type in []
        part = _get_configuration_overwrite_schema_for_input(input)
        if part != None:
            schema["properties"][input["id"]] = part

    return schema

    
def _get_schema_for_input(input: Dict[str, Any]) -> Dict[str, Any]:
    type = input['type']

    match type:
        case 'checkbox' : return { "oneOf" : [ { "type" : "string" }, { "type" : "number" }]}
        case 'date' : return { "type" : "string" }
        case 'display-resource-carousel' : return { "oneOf" : [ { "type" : "string" }, { "type" : "array", "items" : { "type" : "string" }} ] }
        case 'display-resource' : return { "oneOf" : [ { "type" : "string" }, { "type" : "array", "items" : { "type" : "string" }} ] }
        case 'display-text' : return { "type" : "string" }
        case 'display-title' : return { "type" : "string" }
        case 'email' : return { "type" : "string" }
        case 'hidden' : return { "type" : "string" }
        case 'listbox' : return {}
        case 'number' : return { "type" : "number" }
        case 'radio' : return { "oneOf" : [ { "type" : "string" }, { "type" : "number" }]}
        case 'text' : return { "oneOf" : [ { "type" : "string" }, { "type" : "number" }]}
        case 'textarea' : return { "oneOf" : [ { "type" : "string" }, { "type" : "number" }]}
        case 'tool-iat-rectangle' : return _get_iat_with_resource(validation_schemas.LAB_IAT_RECTANGLE_SCHEMA)
        case 'tool-iat-polygon' : return _get_iat_with_resource(validation_schemas.LAB_IAT_SCHEMA)
        case 'tool-iat-polyline' : return _get_iat_with_resource(validation_schemas.LAB_IAT_SCHEMA)
        case 'tool-iat-graph' : return _get_iat_with_resource(validation_schemas.LAB_IAT_SCHEMA)
        case 'tool-iat-segmentation' : return _get_iat_with_resource(validation_schemas.LAB_IAT_SEGMENTATION_SCHEMA)
        case 'url' : return { "oneOf" : [ { "type" : "string" }, { "type" : "number" }]}
        case 'display-json' : return {}
        case 'json-editor' : return {}
    
    return {}
    
    
def _get_configuration_overwrite_schema_for_input(input: Dict[str, Any]) -> Dict[str, Any]:
    type = input['type']

        # "listbox": ['settings.multiple', 'settings.required', 'settings.options'],
        # "radio": ['settings.required', 'settings.options'],
    match type:
        case 'radio' : return { 
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "settings": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "required": {"type": "boolean"},
                        "options": {
                            "type": "array",
                            "items" : {
                                "type": "object",
                                "additionalProperties": False,
                                "properties" : {
                                    "name" : { "type" : "string" },
                                    "value" : { "type" : "string" },
                                    "group" : { "type" : "string" }
                                },
                                "required" : ["name", "value"]
                            }
                        },
                    }
                }
            }
        }
        case 'listbox' : return { 
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "settings": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "multiple": {"type": "boolean"},
                        "required": {"type": "boolean"},
                        "options": {
                            "type": "array",
                            "items" : {
                                "type": "object",
                                "additionalProperties": False,
                                "properties" : {
                                    "name" : { "type" : "string" },
                                    "value" : { "type" : "string" },
                                    "group" : { "type" : "string" }
                                },
                                "required" : ["name", "value"]
                            }
                        },
                    }
                }
            }
        }
    
    return None
    


def _get_iat_with_resource(schema: Dict[str, Any]) -> Dict[str, Any]:
    schema_copy = copy.deepcopy(schema)
    schema_copy["properties"]["resources"] = { "oneOf" : [ { "type" : "string" }, { "type" : "array", "items" : { "type" : "string" }} ] }
    return schema_copy


"""Project domain"""

from typing import NewType, Optional, List, Literal, Dict, Any
from .pagination import PaginationFilters
from dataclasses import dataclass

ProjectId = NewType("ProjectId", str)

ProjectType = Literal["form", "iat-rectangle", "iat-polygon", "iat-segmentation", "iat-graph", "iat-polyline"]
ProjectStatus = Literal["configuring", "pending", "complete", "toBeDeleted", "deleted", "archived"]


@dataclass
class ProjectFilters(PaginationFilters):
    """Project filters for running a project search."""
    id_in: Optional[ProjectId] = None
    status_in: Optional[List[ProjectStatus]] = None
    type_in: Optional[List[ProjectType]] = None
    search: Optional[str] = None



from typing import List, BinaryIO, Dict
from isahitlab.api.base import BaseApi
from ..helpers import get_response_json, log_raise_for_status

class FileApi(BaseApi):
    """File API Calls"""

    def grant_file_access(self, ids: List[str]) -> Dict :
        """Get files grant token"""
        
        access_token = self._http_client.post('api/file-manager/grant/files', json={ "ids": ids })
        
        log_raise_for_status(access_token)
        
        return get_response_json(access_token)
    

    def upload_file(self, batch_id: str, file : BinaryIO, folder) -> Dict :
        """Upload file without dataset"""
        
        files = {'file': file }

        uploaded = self._http_client.post('api/file-manager/files', files=files, params={ "skipTaskCreation" : "true" }, data={ "path": folder, "batchId": batch_id })
        
        log_raise_for_status(uploaded)
        
        return get_response_json(uploaded)


import logging
from typing import List, Optional

from isahitlab.api.task.api import TaskApi
from isahitlab.domain.task import TaskId
from isahitlab.operations.base import BaseAction
from tqdm import tqdm
from typeguard import typechecked

logger = logging.getLogger('isahitlab.to_review_tasks')


class ToReviewTasksOperation(BaseAction):
    """Send tasks to review
    """

    _task_api : TaskApi

    @typechecked
    def run(
        self,
        project_id: str,
        task_id_in: List[TaskId],
        disable_progress_bar: Optional[bool] = False,
    ) -> int:
        """ Send tasks to review
        
        Args:
            project_id: ID of the project
            task_id_in: List of the IDs of the tasks to send to review
            disable_progress_bar: Disable the progress bar display


        """
        self._task_api = TaskApi(self._http_client)
                
        with tqdm(total=len(task_id_in),  disable=disable_progress_bar, desc="Sending tasks to review... ") as loader:
            for task_id in task_id_in:
                self._task_api.to_review_tasks(project_id=project_id, task_id=task_id)
                loader.update(1)
            

        
        return len(task_id_in)
    


"""Integration domain"""

from typing import NewType, Optional, List, Literal
from .pagination import PaginationFilters
from dataclasses import dataclass

IntegrationId = NewType("IntegrationId", str)

IntegrationVisibility = Literal["organization", "private"]
IntegrationType = Literal["GCP", "S3"]


@dataclass
class IntegrationFilters(PaginationFilters):
    """Integration filters."""
    visibility_in: Optional[List[IntegrationVisibility]] = None
    search: Optional[str] = None


@dataclass
class IntegrationPayload:
    """Integration payload
    
    Args:
        name: Name of the batch
        type: Type of the integration,
        visibility: Visibility of the integration
        access_point: Access point (or bucket for GCP)
        role_id: Role ID or ARN for S3 integration
        external_id: Client external ID for S3 integration
    """

    name: str
    type: IntegrationType
    visibility: IntegrationVisibility
    access_point: str
    role_id: Optional[str]
    external_id: Optional[str]



from typing import Dict, List


def bbox_to_corners(bbox: List[float]) -> List[float]:
    """ Transform a list of point representing a bbox to the four points of a rectangle"""

    [top_left_x, top_left_y, bottom_right_x, bottom_right_y] = bbox

    return [
        top_left_x,
        top_left_y,
        bottom_right_x,
        top_left_y,
        bottom_right_x,
        bottom_right_y,
        top_left_x,
        bottom_right_y
    ]

def bbox_to_center_and_size(bbox: List[float]) -> List[float]:
    """ Transform a list of point representing a bbox to the center x and y with width and height"""

    [top_left_x, top_left_y, bottom_right_x, bottom_right_y] = bbox

    min_x = min(top_left_x, bottom_right_x)
    max_x = max(top_left_x, bottom_right_x)
    min_y = min(top_left_y, bottom_right_y)
    max_y = max(top_left_y, bottom_right_y)

    width = max_x - min_x
    height = max_y - min_y

    center_x = min_x + (width) / 2
    center_y = min_y + (height) / 2

    epsilon = 0.01
    # Prevent empty bbox by setting an epsilon size
    if width == 0:
        width = epsilon
    if height == 0:
        height = epsilon

    return [
        center_x,
        center_y,
        width,
        height
    ]

def vertices_to_bbox(vertices: List[float]) -> List[float]:
    """ Transform a list of vertices to a [top_left_x, top_left_y, bottom_right_x, bottom_right_y]"""

    xs = []
    ys = []

    for i, v in enumerate(vertices):
        if i % 2 == 0:
            xs.append(vertices[i])
            ys.append(vertices[i + 1])

    min_x = min(xs)
    max_x = max(xs)
    min_y = min(ys)
    max_y = max(ys)

    return [
        min_x,
        min_y,
        max_x,
        max_y
    ]

def corners_to_bbox(vertices: List[float]) -> List[float]:
    """ Transform a list of point representing a bbox to the four points of a rectangle"""

    [a_x, a_y, b_x, b_y, c_x, c_y, d_x, d_y] = vertices

    
    min_x = min(a_x, b_x, c_x, d_x)
    max_x = max(a_x, b_x, c_x, d_x)
    min_y = min(a_y, b_y, c_y, d_y)
    max_y = max(a_y, b_y, c_y, d_y)

    return [
        min_x,
        min_y,
        max_x,
        max_y
    ]

def vertices_to_points(vertices: List[float]) -> List[Dict]:
    """Transform vertices [float,float,float,float,...] to points [{ x: float, y: float },{ x: float, y: float },...]"""

    points = []

    for i, v in enumerate(vertices):
        if i % 2 == 0:
            points.append({ "x" : vertices[i], "y" : vertices[i + 1]})

    return points

def points_to_vertices(points):
    """Transform points [{ x: float, y: float },{ x: float, y: float },...] to vertices [float,float,float,float,...] """
    vertices = []
    for p in points:
        vertices.append(p['x'])
        vertices.append(p['y'])
    return vertices

def denormalize_points(points: List[Dict], dimension : Dict, round_decimal : int = None) -> List[float]:
    """Transform relative points to absolute points"""

    denormalized_points = []

    for p in points:
        x = p['x'] * float(dimension['width'])
        y = p['y'] * float(dimension['height'])
        if round_decimal != None:
            x = round(x, round_decimal)
            y = round(y, round_decimal)
        denormalized_points.append({"x" : x, "y": y})

    return denormalized_points

def normalize_points(points: List[Dict], dimension : Dict, round_decimal : int = None) -> List[float]:
    """Transform relative points to absolute points"""

    normalized_points = []

    for p in points:
        x = p['x'] / float(dimension['width'])
        y = p['y'] / float(dimension['height'])
        if round_decimal != None:
            x = round(x, round_decimal)
            y = round(y, round_decimal)
        normalized_points.append({"x" : x, "y": y})

    return normalized_points


from typing import Dict, Generator
import time
from isahitlab.api.base import BaseApi
from isahitlab.domain.integration import IntegrationFilters, IntegrationPayload
from ..helpers import get_response_json, log_raise_for_status, Paginated
from .mappers import integration_filters_mapper

class IntegrationApi(BaseApi):
    """Project API Calls"""

    def get_all_integrations(self, filters: IntegrationFilters) -> Generator[tuple[Dict, int, int], None, None] :
        """Get integrations"""

        page = 1
        limit = 10

        filters.pagination = True
        filters.limit = limit

        while True:
            filters.page = page

            paginated: Paginated = self._get_integration_page(filters)
            docs = paginated['docs']
            page = paginated['nextPage']

            # Count already loaded
            totalDocs = paginated['totalDocs']
            loaded = (paginated['page'] - 1) * paginated['limit'] + len(docs)

            yield (docs, loaded, totalDocs)

            if not paginated['hasNextPage']:
                break

            time.sleep(0.5)



    
    def _get_integration_page(self, filters: IntegrationFilters) -> Paginated :
        """Get integration page"""
        
        tasks = self._http_client.get('api/file-manager/integrations', params=integration_filters_mapper(filters))
        
        log_raise_for_status(tasks)
        
        return get_response_json(tasks)
    

    def create_integration(self, integration: IntegrationPayload):
        """Create batch in a project"""

        data = {
            "name": integration.name,
            "type": integration.type,
            "visibility": integration.visibility,
            "accessPoint" : integration.access_point
        }

        if integration.type == 'S3':
            data['roleId'] = integration.role_id
            data['externalId'] = integration.external_id


        result = self._http_client.post('api/file-manager/integrations', json=data, params={ "preventSharedDuplicate" : "true" })

        log_raise_for_status(result)

        return get_response_json(result)



""" Domain Filter to API params"""
from isahitlab.domain.dataset import DatasetBrowsingFilters

def dataset_browsing_filters_mapper(filters: DatasetBrowsingFilters):
    params = {}

    if filters.folder:
        params['folder'] = filters.folder

    if filters.next_token and len(filters.next_token):
        params['nextToken'] = filters.next_token
        
    return params


import logging
from typing import Dict, Iterable, Optional

from isahitlab.api.batch.api import BatchApi
from isahitlab.api.workflow.api import WorkflowApi
from isahitlab.domain.batch import BatchFilters, BatchId
from isahitlab.domain.project import ProjectId
from isahitlab.domain.workflow import WebhookPayload, WorkflowId
from isahitlab.operations.base import BaseAction
from tqdm import tqdm
from typeguard import typechecked

logger = logging.getLogger('isahitlab.setup_webhook')


class SetupWebhook(BaseAction):
    """Create task actions
    """


    _batch_api : BatchApi
    _workflow_api : WorkflowApi

    @typechecked
    def run(
        self,
        webhook: WebhookPayload,
        project_id: Optional[ProjectId] = None,
        batch_id: Optional[BatchId] = None,
        workflow_id: Optional[WorkflowId] = None,
        disable_progress_bar: Optional[bool] = False

    ) -> None:
        """ Setup a webhook for a workflow

        !!! warning
            If you only set the project_id parameter, the webhook will be set on all batches of the project
        
        Args:
            project_id: ID of the project
            batch_id: ID of the batch 
            workflow_id: ID of the workflow project
            webhook : See WebhookPayload (url : str, header : str)
            disable_progress_bar: Disable the progress bar display

        """
        self._batch_api = BatchApi(self._http_client)
        self._workflow_api = WorkflowApi(self._http_client)

        if not project_id and not workflow_id and not batch_id:
            raise ValueError('You must provide either a project_id, a batch_id or a workflow_id')
        
        if workflow_id:
            self._workflow_api.setup_webhook_by_workflow_id(workflow_id=workflow_id, webhook=webhook)
        elif batch_id:
            self._workflow_api.setup_webhook_by_batch_id(batch_id=batch_id, webhook=webhook)
        elif project_id:
            # load batches and apply
            filters : BatchFilters = BatchFilters(project_id=project_id)
            with tqdm(total=0,  disable=disable_progress_bar, desc="Setting up the webhook... ") as loader:
                for (docs, loaded, total) in self._batch_api.get_all_batches(filters):
                    loader.total = total
                    for batch in docs:
                        batch_id = batch['_id']
                        self._workflow_api.setup_webhook_by_batch_id(batch_id=batch_id, webhook=webhook)
                    loader.update(loaded - loader.n)
            


from tqdm import tqdm
from typing import Optional, Dict
from isahitlab.operations.base import BaseAction
from isahitlab.domain.batch import BatchId, BatchPayload
from isahitlab.domain.project import ProjectId
from isahitlab.api.batch.api import BatchApi

from typeguard import typechecked

class CreateBatchOperation(BaseAction):
    """Create batch operation"""

    @typechecked
    def run(
        self,
        project_id: ProjectId,
        batch : BatchPayload,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Create a batch in a project
        
        Args:
            project_id: ID of the batch
            batch : See BatchPayload (name : str)
            disable_progress_bar: Disable the progress bar display
        """

        batch_api = BatchApi(self._http_client)
        with tqdm(total=1,  disable=disable_progress_bar, desc="Creating batch... ") as loader:
            created_batch = batch_api.create_batch(project_id=project_id, batch=batch)
            loader.update(1)

        return created_batch
    


"""Task domain"""
from dataclasses import dataclass
from typing import Optional


@dataclass
class PaginationFilters:
    """Pagination filters."""

    pagination: bool = True
    page: Optional[int] = 1
    limit: Optional[int] = 10
